#include "Win32DxInterface.h"

#include "FileLogger.h"

#include <DirectXMath.h>
#include <DirectXPackedVector.h>

#include <dxgi.h>
#include <dxgi1_6.h>
#include <d3d11.h>
#include <d3d11on12.h>
#include <d3d12.h>
#include <d3d12sdklayers.h>

#include <d2d1_3.h>
#include <d2d1_1.h>

#include <dwrite.h>

#include <iostream>
#include <vector>

#include <windows.h>
#include <wrl.h>

void Win32DxInterface::getHardwareAdapter(IDXGIAdapter1** ppAdapter)
{
	*ppAdapter = nullptr;
	for (UINT adapterIndex = 0; ; ++adapterIndex)
	{
		IDXGIAdapter1* pAdapter = nullptr;
		if (DXGI_ERROR_NOT_FOUND == mFactory->EnumAdapters1(adapterIndex, &pAdapter))
		{
			// No more adapters to enumerate.
			break;
		}

		// Check to see if the adapter supports Direct3D 12, but don't create the actual device yet.
		if (SUCCEEDED(D3D12CreateDevice(pAdapter, D3D_FEATURE_LEVEL_11_0, _uuidof(ID3D12Device), nullptr)))
		{
			*ppAdapter = pAdapter;
			return;
		}
		pAdapter->Release();
	}
}

bool Win32DxInterface::isValid() const
{
	return mIsValid;
}

ID3D12Device* Win32DxInterface::getDevice() const
{
	return mDevice.Get();
}

ID3D11On12Device* Win32DxInterface::get11On12Device() const
{
	return mD3d11On12Device.Get();
}

IDXGIFactory7* Win32DxInterface::getFactory() const
{
	return mFactory.Get();
}

ID2D1Factory3* Win32DxInterface::getD2dFactory() const
{
	return mD2dFactory.Get();
}

IDWriteFactory* Win32DxInterface::getDirectWriteFactory() const
{
	return mDWriteFactory.Get();
}

ID2D1DeviceContext2* Win32DxInterface::getD2dContext() const
{
	return mD2dDeviceContext.Get();
}

ID3D12CommandQueue* Win32DxInterface::getCommandQueue() const
{
	return mCommandQueue.Get();
}

ID3D11DeviceContext* Win32DxInterface::getD3d11DeviceContext() const
{
	return mD3d11DeviceContext.Get();
}

void Win32DxInterface::initialize()
{
	if (!DirectX::XMVerifyCPUSupport())
	{
		MLOG_ERROR("Directx math not supported");
		return;
	}

#if defined(DEBUG) || defined(_DEBUG)
	// Enable the D3D12 debug layer.
	{
		ID3D12Debug* debugController;
		D3D12GetDebugInterface(IID_PPV_ARGS(&debugController));
		debugController->EnableDebugLayer();
	}
#endif

	HRESULT hr = CreateDXGIFactory(__uuidof(IDXGIFactory7), (void**)(&mFactory));

	IDXGIAdapter1* pAdapter{ nullptr };
	getHardwareAdapter(&pAdapter);

	if (pAdapter)
	{
		MLOG_INFO("Found adapter");
		if (SUCCEEDED(D3D12CreateDevice(pAdapter, D3D_FEATURE_LEVEL_11_0, IID_PPV_ARGS(&mDevice))))
		{
			MLOG_INFO("Got device");
		}
	}
	else
	{
		MLOG_INFO("No adapter found");
		return;
	}

	D3D12_COMMAND_QUEUE_DESC queueDesc = {};
	queueDesc.Flags = D3D12_COMMAND_QUEUE_FLAG_NONE;
	queueDesc.Type = D3D12_COMMAND_LIST_TYPE_DIRECT;

	mDevice->CreateCommandQueue(&queueDesc, IID_PPV_ARGS(&mCommandQueue));

	mIsValid = initializeD2d();
}

bool Win32DxInterface::initializeD2d()
{
	Microsoft::WRL::ComPtr<ID3D11Device> d3d11Device;
	UINT d3d11DeviceFlags = D3D11_CREATE_DEVICE_BGRA_SUPPORT;
	D3D11On12CreateDevice(
		mDevice.Get(),
		d3d11DeviceFlags,
		nullptr,
		0,
		reinterpret_cast<IUnknown**>(mCommandQueue.GetAddressOf()),
		1,
		0,
		&d3d11Device,
		&mD3d11DeviceContext,
		nullptr
	);

	d3d11Device.As(&mD3d11On12Device);

	// Create D2D/DWrite components.
	{
		D2D1_DEVICE_CONTEXT_OPTIONS deviceOptions = D2D1_DEVICE_CONTEXT_OPTIONS_NONE;
		D2D1_FACTORY_OPTIONS d2dFactoryOptions = {};

		D2D1CreateFactory(D2D1_FACTORY_TYPE_SINGLE_THREADED, __uuidof(ID2D1Factory3), &d2dFactoryOptions, &mD2dFactory);
		Microsoft::WRL::ComPtr<IDXGIDevice> dxgiDevice;
		mD3d11On12Device.As(&dxgiDevice);

		mD2dFactory->CreateDevice(dxgiDevice.Get(), &mD2dDevice);

		mD2dDevice->CreateDeviceContext(deviceOptions, &mD2dDeviceContext);

		DWriteCreateFactory(DWRITE_FACTORY_TYPE_SHARED, __uuidof(IDWriteFactory), &mDWriteFactory);
	}

	return true;
}